import json
import os
from snowbim.utilities import common
from os.path import expanduser

def load_bim(file_path:str=None):
    '''
    Read .bim file as json object. Then, convert it to dict object
    Return a tuple (code, dict, messasge)
    '''
    bim = None
    if file_path is None or not os.path.exists(file_path):
        return (0, { "model": {} }, None)

    with open(file_path, 'r') as f:
        bim = json.load(f)
    return (0, bim, None)
    

def upgrade_bim(file_path:str=None, out_path:str=None, changes:dict={}):
    '''
    With the changes object, make changes for dict's attributes related.
    Currently support only for
        changes = { "model": { "tables": [ { "columns": [...] }, {...} ] } }
    Return a tuple (code, dict, messasge)
    '''
    input_bim = load_bim(file_path=file_path)
    if input_bim[0] != 0:
        return (input_bim[0], {}, input_bim[2])
    else: 
        input_bim = input_bim[1]

    output_bim = {
        "name": "PowerBITabularModel",
        "compatibilityLevel": 1550,
        "model": {
            "culture": "en-US",
            "dataAccessOptions": {
                "legacyRedirects": "true",
                "returnErrorValuesAsNull": "true"
            },
            "defaultPowerBIDataSourceVersion": "powerBI_V3",
            "sourceQueryCulture": "en-US",
            "tables": [],
            "relationships": [],
            "cultures": [
                {
                    "name": "en-US",
                    "linguisticMetadata": {
                    "content": {
                        "Version": "1.0.0",
                        "Language": "en-US",
                        "DynamicImprovement": "HighConfidence"
                    },
                    "contentType": "json"
                    }
                }
            ],
            "annotations": []
        }
    }

    if not file_path:
        file_path = f'{expanduser("~")}/model.bim'
    if not out_path:
        out_path = file_path
    
    # completely new model
    if not input_bim['model']:
        output_bim['model']['tables'] = changes['model']['tables']
        
    # upgrade model
    else:
        output_bim = input_bim

        in_tables = input_bim['model']['tables']
        for table in changes['model']['tables']:
            change_table_name = table['name']
            in_table = [x for x in in_tables if x['name'] == change_table_name]
            if len(in_table) > 1:
                return (-1, {}, f'Invalid model as input duplicates table: {change_table_name}')

            if len(in_table) == 0:
                # new table
                in_tables.append(table)
            else:
                in_table = in_table[0]

                # existing table
                # existing table \ columns
                in_table_columns = in_table['columns']
                for column in table['columns']:
                    change_colum_name = column['name']
                    in_colum = [x for x in in_table_columns if x['name'] == change_colum_name]
                    if len(in_colum) > 1:
                        return (-1, {}, f'Invalid model as input duplicates column: {change_colum_name}')
                        
                    if len(in_colum) == 0:
                        # existing table \ columns \ new column
                        in_table_columns.append(column)
                    else:
                        in_colum = in_colum[0]
                        # existing table \ columns \ existing column
                        in_colum['dataType'] = column['dataType'] or in_colum['dataType']
                        in_colum['isHidden'] = column['isHidden'] or in_colum['isHidden']
                        in_colum['sourceColumn'] = column['sourceColumn'] or in_colum['sourceColumn']
                
                # existing table \ partitions
                in_table_partitions = in_table['partitions']
                for partition in table['partitions']:
                    in_partition = [
                        x for x in in_table_partitions 
                            if len(set(x['source']['expression']).intersection(partition['source']['expression'])) == len(x['source']['expression'])
                                and x['source']['type'] == partition['source']['type']
                                and x['mode'] == partition['mode']
                    ]
                    
                    if len(in_partition) == 0:
                        # existing table \ partitions \ new partition
                        in_table_partitions.append(partition)
                    else:
                        # existing table \ partitions \ existing partition
                        pass
    print(out_path)
    common.save_to_json_file(output_bim, out_path)
    return (0, output_bim, None)


import json
import datetime

def json_default(o):
    '''
    Json default method
    '''
    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()
    return o


def save_to_json_file(data:dict,file):
    '''
    Save json object to file
    '''
    with open(file, 'w', encoding='utf8') as outfile: 
        outfile.write(json.dumps(data, indent = 2, default=json_default, ensure_ascii=False))

def get_model_datatype(in_type:str=None):
    if in_type is None:
        return 'string'

    mapping = [
        ('text',            'string'),
        ('boolean',         'boolean'),
        ('timestamp_ntz',   'dateTime'),
        ('timestamp_ltz',   'dateTime'),
        ('date',            'dateTime'),
        ('number',          'double')
    ]

    datatype = [x[1] for x in mapping if x[0] == in_type.lower()]
    if datatype is None or len(datatype) == 0:
        return in_type.lower()
    else:
        datatype = datatype[0]

    return datatype
    

def get_model_datatype(in_type:str=None):
    if in_type is None:
        return 'string'

    mapping = [
        ('text',            'string'),
        ('boolean',         'boolean'),
        ('timestamp_ntz',   'dateTime'),
        ('timestamp_ltz',   'dateTime'),
        ('date',            'dateTime'),
        ('number',          'double')
    ]

    datatype = [x[1] for x in mapping if x[0] == in_type.lower()]
    if datatype is None or len(datatype) == 0:
        return in_type.lower()
    else:
        datatype = datatype[0]

    return datatype


import os
import json
from os.path import expanduser
import yaml
import snowflake.connector as sfconn
from yaml.loader import FullLoader

from snowbim.utilities import common


def connect(profile_dir:str=None, profile:str=None, target:str=None, db:str=None, schema:str=None):
    '''
    Use dbt profiles.yml to connect snowflake database
    Returns a tuple (code, snowflake connection, message)
    '''
    conn = {}

    profile_path = ''
    if not profile_dir:
        profile_path = f'{expanduser("~")}/.dbt/profiles.yml'
    else:
        profile_path = f'{profile_dir}/profiles.yml'
    print(f'dbt profile was found at: {profile_path}')

    with open(profile_path, 'r') as stream:
        try:
            cred = yaml.load(stream, Loader=FullLoader)
        except yaml.YAMLError as e:
            return (-1, {}, str(e))

    if profile is None:
        cred = cred[next(iter(cred))]
    else:
        cred = cred[profile]

    cred = cred['outputs']
    if target is None:
        cred = cred[next(iter(cred))]
    else:
        cred = cred[target]

    if cred is None or cred['type'] != 'snowflake':
        return (-1, {}, f'Snowflake config not found: project={str(profile)}, target={str(target)} within profile: {profile_path}')

    # cred = {'type': 'snowflake', 'account': 'xxx', 'user': 'xxx', 'password': 'xxx', 'role': 'xxx', 'database': 'xxx', 'warehouse': 'xxx', 'schema': 'xxx'}
    conn = sfconn.connect(
        user = cred['user'],
        password = cred['password'],
        account = cred['account'],
        warehouse = cred['warehouse'] or 'COMPUTE_WH',
        database = db or cred['database'] or 'DEMO_DB',
        schema = schema or cred['schema'] or 'PUBLIC',
        role = cred['role'] or 'PUBLIC'
    )
    
    return (0, conn, None)


def compare_schema(snowflake_conn, bim_path:str=None, mode:str='directQuery', table_excludes:list=[]):
    '''
    Get the changes of snowflake database schema
    Currently support only for
        changes = { "model": { "tables": [ { "columns": [...] }, {...} ] } }
    Returns a tuple (code, changes, message)
    '''
    # Input bim data
    in_schema = {}
    if bim_path and os.path.exists(bim_path):
        with open(bim_path, 'r') as f:
            in_schema = json.load(f)
            in_schema = in_schema['model']['tables']

    # SF bim data
    snowflake_schema = []
    cur = snowflake_conn.cursor()

    cur.execute(f'SELECT * FROM "INFORMATION_SCHEMA"."TABLES" WHERE "TABLE_SCHEMA" = \'{snowflake_conn.schema}\' ORDER BY "TABLE_SCHEMA", "TABLE_NAME"')
    df_tables = cur.fetch_pandas_all()

    cur.execute(f'SELECT * FROM "INFORMATION_SCHEMA"."COLUMNS" WHERE "TABLE_SCHEMA" = \'{snowflake_conn.schema}\' ORDER BY "TABLE_SCHEMA", "TABLE_NAME", "COLUMN_NAME"')
    df_columns = cur.fetch_pandas_all()

    for index, item in df_tables.iterrows():
        table_item = {
            "name": item['TABLE_NAME'],
            "is_new": 1,
            "columns": [],
            "partitions": [
                {
                    "name": f"{item['TABLE_NAME']} Partition",
                    "is_new": 1,
                    "mode": mode,
                    "source": {
                        "type": "m",
                        "expression": [
                            f"let",
                            f"    Source = Snowflake.Databases(\"{snowflake_conn.account}.snowflakecomputing.com\", \"{snowflake_conn.warehouse}\", [Role=\"{snowflake_conn.role}\", CreateNavigationProperties=null, ConnectionTimeout=null, CommandTimeout=null]),",
                            f"    {snowflake_conn.database}_Database = Source{{[Name=\"{snowflake_conn.database}\",Kind=\"Database\"]}}[Data],",
                            f"    {snowflake_conn.schema}_Schema = {snowflake_conn.database}_Database{{[Name=\"{snowflake_conn.schema}\",Kind=\"Schema\"]}}[Data],",
                            f"    #\"{item['TABLE_NAME']}_{item['TABLE_TYPE'].title()}\" = {snowflake_conn.schema}_Schema{{[Name=\"{item['TABLE_NAME']}\",Kind=\"{item['TABLE_TYPE'].title()}\"]}}[Data]",
                            f"in",
                            f"    #\"{item['TABLE_NAME']}_{item['TABLE_TYPE'].title()}\""
                        ]
                    }
                }
            ]
        }

        df_filterred_columns = df_columns.loc[df_columns['TABLE_NAME'] == item['TABLE_NAME']]
        for index, citem in df_filterred_columns.iterrows():
            table_item['columns'].append({
                "name": citem['COLUMN_NAME'],
                "is_new": 1,
                "dataType": citem['DATA_TYPE'],
                "sourceColumn": citem['COLUMN_NAME']
            })
        snowflake_schema.append(table_item)
    
    # Detect changes
    changes = { "model": { "tables": [ ] } }
    for sftable in snowflake_schema:
        in_table = [x for x in in_schema if x['name'] == sftable['name']]
        if in_table:
            # existing table
            in_table = in_table[0]

            table = {}
            table['name'] = in_table['name']
            table['is_new'] = 0

            # columns
            table['columns'] = []
            for sfcolumn in sftable['columns']:
                in_column = [x for x in in_table['columns'] if x['name'] == sfcolumn['name']]
                if in_column:
                    # existing column
                    in_column = in_column[0]
                    sf_model_type = common.get_model_datatype(sfcolumn['dataType'])
                    if (in_column['dataType'] != sf_model_type or in_column['sourceColumn'] != sfcolumn['sourceColumn']):
                        in_column['dataType'] = sf_model_type
                        in_column['sourceColumn'] = sfcolumn['sourceColumn']
                        in_column['is_new'] = 0

                        table['columns'].append(in_column)
                else:
                    # new column
                    sfcolumn['dataType'] = common.get_model_datatype(sfcolumn['dataType'])
                    table['columns'].append(sfcolumn)

            # partitions
            table['partitions'] = []
            for sfpartition in sftable['partitions']:
                in_partition = [
                    x for x in in_table['partitions']
                        if len(set(x['source']['expression']).intersection(sfpartition['source']['expression'])) == len(x['source']['expression'])
                            and x['source']['type'] == sfpartition['source']['type']
                            and x['mode'] == sfpartition['mode']
                ]
                if in_partition:
                    # existing partition
                    pass
                else:
                    # new partition
                    table['partitions'].append(sfpartition)

            if len(table['columns']) > 0 or len(table['partitions']) > 0:
                changes['model']['tables'].append(table)
        else:
            # new table
            for sfcolumn in sftable['columns']:
                sfcolumn['dataType'] = common.get_model_datatype(sfcolumn['dataType'])
            changes['model']['tables'].append(sftable)
            
    return (0, changes, None)


from snowbim.engines import snowengine, bimengine

def get_schema_changes(bim_path=None, profile_dir:str=None, profile:str=None, target:str=None, db:str=None, schema:str=None):
    '''
    get_schema_changes
    '''
    print('Connecting to Snowflake...')
    conn = snowengine.connect(profile_dir=profile_dir, profile=profile, target=target, db=db, schema=schema)
    print('     Connected')

    print('Schema comparing...')
    changes = snowengine.compare_schema(snowflake_conn=conn[1], bim_path=bim_path)
    print('     Done')

    return changes

def upgrade_schema(bim_path=None, out_bim_path=None, profile_dir:str=None, profile:str=None, target:str=None, db:str=None, schema:str=None):
    '''
    upgrade_schema
    '''
    changes = get_schema_changes(bim_path=bim_path, profile_dir=profile_dir, profile=profile, target=target, db=db, schema=schema)
    if changes[0] == 0:
        changes = changes[1]
    else:
        return (changes[0], {}, changes[2])

    print(f"Changes detected as: { 'Up-to-date' if len(changes['model']['tables']) == 0 else ''}")
    for table in changes['model']['tables']:
        print(f"    Table: {table['name']} {'(New)' if table['is_new'] else '(Changed)'}")
        del table['is_new']
        for column in table['columns']:
            print(f"        Column: {column['name']} {'(New)' if column['is_new'] else '(Changed)'}")
            del column['is_new']
        for partition in table['partitions']:
            print(f"        Partition: {partition['name']} {'(New)' if partition['is_new'] else '(Changed)'}")
            del partition['is_new']

    if len(changes['model']['tables']):
        print('Upgrading...')
        r = bimengine.upgrade_bim(file_path=bim_path, out_path=out_bim_path, changes=changes)
        print('     Done')
        return r

    return (0,{},None)


import argparse, sys
from snowbim import snowbim

def main():
    parser = argparse.ArgumentParser(prog='snowbim')
    parser.add_argument('-v','--version', action='version', version='%(prog)s 1.0.9')

    parser.add_argument('--bim',            help='File path to an input .bim file', type=str, default='')
    parser.add_argument('--out',            help='File path to an out .bim file. Optional', type=str, default='')
    parser.add_argument('--profile_dir',    help='File path to dbt profiles.yml. Optional, default to .dbt folder within User Home one', type=str, default=None)
    parser.add_argument('--profile',        help='dbt profile name. Optional, default: Fisrt profile in profiles.yml', type=str, default=None)
    parser.add_argument('--target',         help='dbt profile target. Optional, default: dev', type=str, default='dev')
    parser.add_argument('--db',             help='Snowflake database. Optional, default: DEMO_DB', type=str, default='DEMO_DB')
    parser.add_argument('--schema',         help='Snowflake database\'s schema. Optional, default: PUBLIC', type=str, default='PUBLIC')

    args = parser.parse_args()
    snowbim.upgrade_schema( bim_path        = args.bim, 
                            out_bim_path    = args.out,
                            profile_dir     = args.profile_dir,
                            profile         = args.profile,
                            target          = args.target,
                            db              = args.db,
                            schema          = args.schema)

if __name__ == '__main__':
    main()


from setuptools import setup, find_packages

setup(
     name='snowbim',
     version='1.0.9',
     author='datnguye',
     author_email='datnguyen.it09@gmail.com',
     packages=find_packages(),
     url='https://github.com/datnguye/snowbim',
     license='MIT',
     description='A package to refresh power bi model .bim file',
     long_description_content_type="text/markdown",
     long_description=open('README.md').read(),
     install_requires=[
        'pyyaml==5.4.1',
        'snowflake-connector-python==2.4.3',
        'snowflake-connector-python[pandas]'
     ],
     python_requires='>=3.7.5',
     entry_points = {
        'console_scripts': [
            'snowbim = snowbim.__main__:main'
        ],
    }
)